@@ -34,15 +34,15 @@ static char *api_extensions[] = {
 	"network_gateway_device_route",
 	"network_phys_macvlan_mtu",
 	"network_veth_router",
-#ifdef HAVE_STRUCT_BPF_CGROUP_DEV_CTX
 	"cgroup2_devices",
-#endif
 	"cgroup2",
 	"pidfd",
 	"seccomp_allow_deny_syntax",
 	"devpts_fd",
 	"seccomp_notify_fd_active",
 	"seccomp_proxy_send_notify_fd",
+	"idmapped_mounts",
+	"idmapped_mounts_v2",
 };
 
 static size_t nr_api_extensions = sizeof(api_extensions) / sizeof(*api_extensions);
