@@ -165,11 +165,10 @@ struct lxc_netdev *lxc_network_add(struct lxc_list *networks, int idx, bool tail
 	__do_free struct lxc_netdev *netdev = NULL;
 
 	/* network does not exist */
-	netdev = malloc(sizeof(*netdev));
+	netdev = zalloc(sizeof(*netdev));
 	if (!netdev)
 		return ret_set_errno(NULL, ENOMEM);
 
-	memset(netdev, 0, sizeof(*netdev));
 	lxc_list_init(&netdev->ipv4);
 	lxc_list_init(&netdev->ipv6);
 
@@ -177,11 +176,9 @@ struct lxc_netdev *lxc_network_add(struct lxc_list *networks, int idx, bool tail
 	netdev->idx = idx;
 
 	/* prepare new list */
-	newlist = malloc(sizeof(*newlist));
+	newlist = lxc_list_new();
 	if (!newlist)
 		return ret_set_errno(NULL, ENOMEM);
-
-	lxc_list_init(newlist);
 	newlist->elem = netdev;
 
 	if (tail)
@@ -199,25 +196,27 @@ struct lxc_netdev *lxc_network_add(struct lxc_list *networks, int idx, bool tail
 struct lxc_netdev *lxc_get_netdev_by_idx(struct lxc_conf *conf,
 					 unsigned int idx, bool allocate)
 {
-	struct lxc_netdev *netdev = NULL;
 	struct lxc_list *networks = &conf->network;
 	struct lxc_list *insert = networks;
 
 	/* lookup network */
 	if (!lxc_list_empty(networks)) {
 		lxc_list_for_each(insert, networks) {
-			netdev = insert->elem;
+			struct lxc_netdev *netdev = insert->elem;
+
+			/* found network device */
 			if (netdev->idx == idx)
 				return netdev;
-			else if (netdev->idx > idx)
+
+			if (netdev->idx > idx)
 				break;
 		}
 	}
 
-	if (!allocate)
-		return ret_set_errno(NULL, EINVAL);
+	if (allocate)
+		return lxc_network_add(insert, idx, true);
 
-	return lxc_network_add(insert, idx, true);
+	return NULL;
 }
 
 void lxc_log_configured_netdevs(const struct lxc_conf *conf)
@@ -225,8 +224,7 @@ void lxc_log_configured_netdevs(const struct lxc_conf *conf)
 	struct lxc_netdev *netdev;
 	struct lxc_list *it = (struct lxc_list *)&conf->network;;
 
-	if ((conf->loglevel != LXC_LOG_LEVEL_TRACE) &&
-	    (lxc_log_get_level() != LXC_LOG_LEVEL_TRACE))
+	if (!lxc_log_trace())
 		return;
 
 	if (lxc_list_empty(it)) {
@@ -248,6 +246,7 @@ void lxc_log_configured_netdevs(const struct lxc_conf *conf)
 		switch (netdev->type) {
 		case LXC_NET_VETH:
 			TRACE("type: veth");
+			TRACE("veth mode: %d", netdev->priv.veth_attr.mode);
 
 			if (netdev->priv.veth_attr.pair[0] != '\0')
 				TRACE("veth pair: %s",
@@ -260,6 +259,15 @@ void lxc_log_configured_netdevs(const struct lxc_conf *conf)
 			if (netdev->priv.veth_attr.ifindex > 0)
 				TRACE("host side ifindex for veth device: %d",
 				      netdev->priv.veth_attr.ifindex);
+
+			if (netdev->priv.veth_attr.vlan_id_set)
+				TRACE("veth vlan id: %d", netdev->priv.veth_attr.vlan_id);
+
+			lxc_list_for_each_safe(cur, &netdev->priv.veth_attr.vlan_tagged_ids, next) {
+				unsigned short vlan_tagged_id = PTR_TO_USHORT(cur->elem);
+				TRACE("veth vlan tagged id: %u", vlan_tagged_id);
+			}
+
 			break;
 		case LXC_NET_MACVLAN:
 			TRACE("type: macvlan");
@@ -395,26 +403,29 @@ void lxc_log_configured_netdevs(const struct lxc_conf *conf)
 	}
 }
 
-static void lxc_free_netdev(struct lxc_netdev *netdev)
+void lxc_clear_netdev(struct lxc_netdev *netdev)
 {
 	struct lxc_list *cur, *next;
+	ssize_t idx;
 
 	if (!netdev)
 		return;
 
-	free(netdev->upscript);
-	free(netdev->downscript);
-	free(netdev->hwaddr);
-	free(netdev->mtu);
+	idx = netdev->idx;
 
-	free(netdev->ipv4_gateway);
+	free_disarm(netdev->upscript);
+	free_disarm(netdev->downscript);
+	free_disarm(netdev->hwaddr);
+	free_disarm(netdev->mtu);
+
+	free_disarm(netdev->ipv4_gateway);
 	lxc_list_for_each_safe(cur, &netdev->ipv4, next) {
 		lxc_list_del(cur);
 		free(cur->elem);
 		free(cur);
 	}
 
-	free(netdev->ipv6_gateway);
+	free_disarm(netdev->ipv6_gateway);
 	lxc_list_for_each_safe(cur, &netdev->ipv6, next) {
 		lxc_list_del(cur);
 		free(cur->elem);
@@ -433,24 +444,43 @@ static void lxc_free_netdev(struct lxc_netdev *netdev)
 			free(cur->elem);
 			free(cur);
 		}
+
+		lxc_list_for_each_safe(cur, &netdev->priv.veth_attr.vlan_tagged_ids, next) {
+			lxc_list_del(cur);
+			free(cur);
+		}
 	}
 
-	free(netdev);
+	memset(netdev, 0, sizeof(struct lxc_netdev));
+	lxc_list_init(&netdev->ipv4);
+	lxc_list_init(&netdev->ipv6);
+	netdev->type = -1;
+	netdev->idx = idx;
 }
 
-define_cleanup_function(struct lxc_netdev *, lxc_free_netdev);
+static void lxc_free_netdev(struct lxc_netdev *netdev)
+{
+	if (netdev) {
+		lxc_clear_netdev(netdev);
+		free(netdev);
+	}
+}
 
 bool lxc_remove_nic_by_idx(struct lxc_conf *conf, unsigned int idx)
 {
-	call_cleaner(lxc_free_netdev) struct lxc_netdev *netdev = NULL;
 	struct lxc_list *cur, *next;
 
+	if (lxc_list_empty(&conf->network))
+		return false;
+
 	lxc_list_for_each_safe(cur, &conf->network, next) {
-		netdev = cur->elem;
+		struct lxc_netdev *netdev = cur->elem;
+
 		if (netdev->idx != idx)
 			continue;
 
 		lxc_list_del(cur);
+		lxc_free_netdev(netdev);
 		free(cur);
 		return true;
 	}
@@ -464,7 +494,8 @@ void lxc_free_networks(struct lxc_list *networks)
 
 	lxc_list_for_each_safe (cur, networks, next) {
 		struct lxc_netdev *netdev = cur->elem;
-		netdev = cur->elem;
+
+		lxc_list_del(cur);
 		lxc_free_netdev(netdev);
 		free(cur);
 	}
@@ -484,7 +515,7 @@ static struct lxc_veth_mode {
 int lxc_veth_mode_to_flag(int *mode, const char *value)
 {
 	for (size_t i = 0; i < sizeof(veth_mode) / sizeof(veth_mode[0]); i++) {
-		if (strcmp(veth_mode[i].name, value) != 0)
+		if (!strequal(veth_mode[i].name, value))
 			continue;
 
 		*mode = veth_mode[i].mode;
@@ -519,7 +550,7 @@ static struct lxc_macvlan_mode {
 int lxc_macvlan_mode_to_flag(int *mode, const char *value)
 {
 	for (size_t i = 0; i < sizeof(macvlan_mode) / sizeof(macvlan_mode[0]); i++) {
-		if (strcmp(macvlan_mode[i].name, value))
+		if (!strequal(macvlan_mode[i].name, value))
 			continue;
 
 		*mode = macvlan_mode[i].mode;
@@ -553,7 +584,7 @@ static struct lxc_ipvlan_mode {
 int lxc_ipvlan_mode_to_flag(int *mode, const char *value)
 {
 	for (size_t i = 0; i < sizeof(ipvlan_mode) / sizeof(ipvlan_mode[0]); i++) {
-		if (strcmp(ipvlan_mode[i].name, value) != 0)
+		if (!strequal(ipvlan_mode[i].name, value))
 			continue;
 
 		*mode = ipvlan_mode[i].mode;
@@ -587,7 +618,7 @@ static struct lxc_ipvlan_isolation {
 int lxc_ipvlan_isolation_to_flag(int *flag, const char *value)
 {
 	for (size_t i = 0; i < sizeof(ipvlan_isolation) / sizeof(ipvlan_isolation[0]); i++) {
-		if (strcmp(ipvlan_isolation[i].name, value) != 0)
+		if (!strequal(ipvlan_isolation[i].name, value))
 			continue;
 
 		*flag = ipvlan_isolation[i].flag;
@@ -636,7 +667,13 @@ int set_config_string_item_max(char **conf_item, const char *value, size_t max)
 
 int set_config_path_item(char **conf_item, const char *value)
 {
-	return set_config_string_item_max(conf_item, value, PATH_MAX);
+	__do_free char *valdup = NULL;
+
+	valdup = path_simplify(value);
+	if (!valdup)
+		return -ENOMEM;
+
+	return set_config_string_item_max(conf_item, valdup, PATH_MAX);
 }
 
 int set_config_bool_item(bool *conf_item, const char *value, bool empty_conf_action)
@@ -698,18 +735,18 @@ bool lxc_config_net_is_hwaddr(const char *line)
 	unsigned index;
 	char tmp[7];
 
-	if (strncmp(line, "lxc.net", 7) != 0)
+	if (!strnequal(line, "lxc.net", 7))
 		return false;
 
-	if (strncmp(line, "lxc.net.hwaddr", 14) == 0)
+	if (strnequal(line, "lxc.net.hwaddr", 14))
 		return true;
 
-	if (strncmp(line, "lxc.network.hwaddr", 18) == 0)
+	if (strnequal(line, "lxc.network.hwaddr", 18))
 		return true;
 
 	if (sscanf(line, "lxc.net.%u.%6s", &index, tmp) == 2 ||
 	    sscanf(line, "lxc.network.%u.%6s", &index, tmp) == 2)
-		return strncmp(tmp, "hwaddr", 6) == 0;
+		return strnequal(tmp, "hwaddr", 6);
 
 	return false;
 }
@@ -754,18 +791,17 @@ bool new_hwaddr(char *hwaddr)
 
 	seed = randseed(false);
 
-	ret = snprintf(hwaddr, 18, "00:16:3e:%02x:%02x:%02x", rand_r(&seed) % 255,
+	ret = strnprintf(hwaddr, 18, "00:16:3e:%02x:%02x:%02x", rand_r(&seed) % 255,
 		       rand_r(&seed) % 255, rand_r(&seed) % 255);
 #else
 
 	(void)randseed(true);
 
-	ret = snprintf(hwaddr, 18, "00:16:3e:%02x:%02x:%02x", rand() % 255,
+	ret = strnprintf(hwaddr, 18, "00:16:3e:%02x:%02x:%02x", rand() % 255,
 		       rand() % 255, rand() % 255);
 #endif
-	if (ret < 0 || ret >= 18) {
-		return log_error_errno(false, EIO, "Failed to call snprintf()");
-	}
+	if (ret < 0)
+		return log_error_errno(false, EIO, "Failed to call strnprintf()");
 
 	return true;
 }
@@ -1006,35 +1042,48 @@ static int sig_num(const char *sig)
 
 static int rt_sig_num(const char *signame)
 {
-	int rtmax = 0, sig_n = 0;
+	bool rtmax;
+	int sig_n = 0;
 
-	if (strncasecmp(signame, "max-", 4) == 0)
-		rtmax = 1;
+	if (is_empty_string(signame))
+		return ret_errno(EINVAL);
 
-	signame += 4;
-	if (!isdigit(*signame))
+	if (strncasecmp(signame, "max-", STRLITERALLEN("max-")) == 0) {
+		rtmax = true;
+		signame += STRLITERALLEN("max-");
+	} else if (strncasecmp(signame, "min+", STRLITERALLEN("min+")) == 0) {
+		rtmax = false;
+		signame += STRLITERALLEN("min+");
+	} else {
+		return ret_errno(EINVAL);
+	}
+
+	if (is_empty_string(signame) || !isdigit(*signame))
 		return ret_errno(EINVAL);
 
 	sig_n = sig_num(signame);
-	sig_n = rtmax ? SIGRTMAX - sig_n : SIGRTMIN + sig_n;
-	if (sig_n > SIGRTMAX || sig_n < SIGRTMIN)
+	if (sig_n < 0 || sig_n > SIGRTMAX - SIGRTMIN)
 		return ret_errno(EINVAL);
 
+	if (rtmax)
+		sig_n = SIGRTMAX - sig_n;
+	else
+		sig_n = SIGRTMIN + sig_n;
+
 	return sig_n;
 }
 
 int sig_parse(const char *signame)
 {
-	size_t n;
-
-	if (isdigit(*signame)) {
+	if (isdigit(*signame))
 		return sig_num(signame);
-	} else if (strncasecmp(signame, "sig", 3) == 0) {
-		signame += 3;
-		if (strncasecmp(signame, "rt", 2) == 0)
-			return rt_sig_num(signame + 2);
 
-		for (n = 0; n < sizeof(signames) / sizeof((signames)[0]); n++)
+	if (strncasecmp(signame, "sig", STRLITERALLEN("sig")) == 0) {
+		signame += STRLITERALLEN("sig");
+		if (strncasecmp(signame, "rt", STRLITERALLEN("rt")) == 0)
+			return rt_sig_num(signame + STRLITERALLEN("rt"));
+
+		for (size_t n = 0; n < ARRAY_SIZE(signames); n++)
 			if (strcasecmp(signames[n].name, signame) == 0)
 				return signames[n].num;
 	}
