@@ -22,6 +22,7 @@
 #include "log.h"
 #include "lxc.h"
 #include "lxclock.h"
+#include "memory_utils.h"
 #include "network.h"
 #include "storage.h"
 #include "syscall_wrappers.h"
@@ -82,8 +83,8 @@ static int load_tty_major_minor(char *directory, char *output, int len)
 	char path[PATH_MAX];
 	ssize_t ret;
 
-	ret = snprintf(path, sizeof(path), "%s/tty.info", directory);
-	if (ret < 0 || (size_t)ret >= sizeof(path))
+	ret = strnprintf(path, sizeof(path), "%s/tty.info", directory);
+	if (ret < 0)
 		return ret_errno(EIO);
 
 	ret = lxc_read_from_file(path, output, len);
@@ -143,18 +144,34 @@ static int cmp_version(const char *v1, const char *v2)
 	return -1;
 }
 
-static void exec_criu(struct cgroup_ops *cgroup_ops, struct lxc_conf *conf,
-		      struct criu_opts *opts)
+struct criu_exec_args {
+	int argc;
+	char *argv[];
+};
+
+static void put_criu_exec_args(struct criu_exec_args *args)
 {
-	char **argv, log[PATH_MAX];
-	int static_args = 23, argc = 0, i, ret;
+	if (args) {
+		for (int i = 0; i < args->argc; i++)
+			free_disarm(args->argv[i]);
+		free_disarm(args);
+	}
+}
+
+define_cleanup_function(struct criu_exec_args *, put_criu_exec_args);
+
+static int exec_criu(struct cgroup_ops *cgroup_ops, struct lxc_conf *conf,
+		     struct criu_opts *opts)
+{
+	call_cleaner(put_criu_exec_args) struct criu_exec_args *args = NULL;
+	__do_fclose FILE *f_mnt = NULL;
+	char log[PATH_MAX];
+	int static_args = 23, ret;
 	int netnr = 0;
 	struct lxc_list *it;
-	FILE *mnts;
 	struct mntent mntent;
 
 	char buf[4096], ttys[32];
-	size_t pos;
 
 	/* If we are currently in a cgroup /foo/bar, and the container is in a
 	 * cgroup /lxc/foo, lxcfs will give us an ENOENT if some task in the
@@ -163,10 +180,8 @@ static void exec_criu(struct cgroup_ops *cgroup_ops, struct lxc_conf *conf,
 	 * /actual/ root cgroup so that lxcfs thinks criu has enough rights to
 	 * see all cgroups.
 	 */
-	if (!cgroup_ops->escape(cgroup_ops, conf)) {
-		ERROR("failed to escape cgroups");
-		return;
-	}
+	if (!cgroup_ops->criu_escape(cgroup_ops, conf))
+		return log_error_errno(-ENOENT, ENOENT, "Failed to escape to root cgroup");
 
 	/* The command line always looks like:
 	 * criu $(action) --tcp-established --file-locks --link-remap \
@@ -176,7 +191,7 @@ static void exec_criu(struct cgroup_ops *cgroup_ops, struct lxc_conf *conf,
 	 * --enable-fs hugetlbfs --enable-fs tracefs --ext-mount-map console:/dev/pts/n
 	 * +1 for final NULL */
 
-	if (strcmp(opts->action, "dump") == 0 || strcmp(opts->action, "pre-dump") == 0) {
+	if (strequal(opts->action, "dump") || strequal(opts->action, "pre-dump")) {
 		/* -t pid --freeze-cgroup /lxc/ct */
 		static_args += 4;
 
@@ -189,7 +204,7 @@ static void exec_criu(struct cgroup_ops *cgroup_ops, struct lxc_conf *conf,
 			static_args += 5;
 
 		/* --leave-running (only for final dump) */
-		if (strcmp(opts->action, "dump") == 0 && !opts->user->stop)
+		if (strequal(opts->action, "dump") && !opts->user->stop)
 			static_args++;
 
 		/* --external tty[88,4] */
@@ -203,7 +218,7 @@ static void exec_criu(struct cgroup_ops *cgroup_ops, struct lxc_conf *conf,
 		/* --ghost-limit 1024 */
 		if (opts->user->ghost_limit)
 			static_args += 2;
-	} else if (strcmp(opts->action, "restore") == 0) {
+	} else if (strequal(opts->action, "restore")) {
 		/* --root $(lxc_mount_point) --restore-detached
 		 * --restore-sibling
 		 * --lsm-profile apparmor:whatever
@@ -212,17 +227,19 @@ static void exec_criu(struct cgroup_ops *cgroup_ops, struct lxc_conf *conf,
 
 		ttys[0] = 0;
 		if (load_tty_major_minor(opts->user->directory, ttys, sizeof(ttys)))
-			return;
+			return log_error_errno(-EINVAL, EINVAL, "Failed to load tty information");
 
 		/* --inherit-fd fd[%d]:tty[%s] */
 		if (ttys[0])
 			static_args += 2;
+
+		static_args += lxc_list_len(&opts->c->lxc_conf->network) * 2;
 	} else {
-		return;
+		return log_error_errno(-EINVAL, EINVAL, "Invalid criu operation specified");
 	}
 
-	if (cgroup_ops->num_hierarchies(cgroup_ops) > 0)
-		static_args += 2 * cgroup_ops->num_hierarchies(cgroup_ops);
+	if (cgroup_ops->criu_num_hierarchies(cgroup_ops) > 0)
+		static_args += 2 * cgroup_ops->criu_num_hierarchies(cgroup_ops);
 
 	if (opts->user->verbose)
 		static_args++;
@@ -232,34 +249,28 @@ static void exec_criu(struct cgroup_ops *cgroup_ops, struct lxc_conf *conf,
 
 	static_args += 2 * lxc_list_len(&opts->c->lxc_conf->mount_list);
 
-	ret = snprintf(log, PATH_MAX, "%s/%s.log", opts->user->directory, opts->action);
-	if (ret < 0 || ret >= PATH_MAX) {
-		ERROR("logfile name too long");
-		return;
-	}
+	ret = strnprintf(log, sizeof(log), "%s/%s.log", opts->user->directory, opts->action);
+	if (ret < 0)
+		return ret_errno(EIO);
 
-	argv = malloc(static_args * sizeof(*argv));
-	if (!argv)
-		return;
+	args = zalloc(sizeof(struct criu_exec_args) + (static_args * sizeof(char **)));
+	if (!args)
+		return log_error_errno(-ENOMEM, ENOMEM, "Failed to allocate static arguments");
 
-	memset(argv, 0, static_args * sizeof(*argv));
-
-#define DECLARE_ARG(arg) 					\
-	do {							\
-		if (arg == NULL) {				\
-			ERROR("Got NULL argument for criu");	\
-			goto err;				\
-		}						\
-		argv[argc++] = strdup(arg);			\
-		if (!argv[argc-1])				\
-			goto err;				\
+#define DECLARE_ARG(arg)                                                                 \
+	do {                                                                             \
+		if (arg == NULL)                                                         \
+			return log_error_errno(-EINVAL, EINVAL,                          \
+					       "Got NULL argument for criu");            \
+		args->argv[(args->argc)++] = strdup(arg);                                \
+		if (!args->argv[args->argc - 1])                                         \
+			return log_error_errno(-ENOMEM, ENOMEM,				 \
+					"Failed to duplicate argumen %s", arg);          \
 	} while (0)
 
-	argv[argc++] = on_path("criu", NULL);
-	if (!argv[argc-1]) {
-		ERROR("Couldn't find criu binary");
-		goto err;
-	}
+	args->argv[(args->argc)++] = on_path("criu", NULL);
+	if (!args->argv[args->argc - 1])
+		return log_error_errno(-ENOENT, ENOENT, "Failed to find criu binary");
 
 	DECLARE_ARG(opts->action);
 	DECLARE_ARG("--tcp-established");
@@ -279,65 +290,52 @@ static void exec_criu(struct cgroup_ops *cgroup_ops, struct lxc_conf *conf,
 	DECLARE_ARG("-o");
 	DECLARE_ARG(log);
 
-	for (i = 0; i < cgroup_ops->num_hierarchies(cgroup_ops); i++) {
-		char **controllers = NULL, *fullname;
-		char *path, *tmp;
+	for (int i = 0; i < cgroup_ops->criu_num_hierarchies(cgroup_ops); i++) {
+		__do_free char *cgroup_base_path = NULL, *controllers;
+		char **controllers_list = NULL;
+		char *tmp;
 
-		if (!cgroup_ops->get_hierarchies(cgroup_ops, i, &controllers)) {
-			ERROR("failed to get hierarchy %d", i);
-			goto err;
-		}
+		if (!cgroup_ops->criu_get_hierarchies(cgroup_ops, i, &controllers_list))
+			return log_error_errno(-ENOENT, ENOENT, "Failed to retrieve cgroup hierarchies %d", i);
 
-		/* if we are in a dump, we have to ask the monitor process what
+		/*
+		 * If we are in a dump, we have to ask the monitor process what
 		 * the right cgroup is. if this is a restore, we can just use
 		 * the handler the restore task created.
 		 */
-		if (!strcmp(opts->action, "dump") || !strcmp(opts->action, "pre-dump")) {
-			path = lxc_cmd_get_limiting_cgroup_path(opts->c->name, opts->c->config_path, controllers[0]);
-			if (!path) {
-				ERROR("failed to get cgroup path for %s", controllers[0]);
-				goto err;
-			}
+		if (strequal(opts->action, "dump") || strequal(opts->action, "pre-dump")) {
+			cgroup_base_path = lxc_cmd_get_limit_cgroup_path(opts->c->name, opts->c->config_path, controllers_list[0]);
+			if (!cgroup_base_path)
+				return log_error_errno(-ENOENT, ENOENT, "Failed to retrieve limit cgroup path for %s", controllers_list[0] ?: "(null)");
 		} else {
 			const char *p;
 
-			p = cgroup_ops->get_limiting_cgroup(cgroup_ops, controllers[0]);
-			if (!p) {
-				ERROR("failed to get cgroup path for %s", controllers[0]);
-				goto err;
-			}
+			p = cgroup_ops->get_limit_cgroup(cgroup_ops, controllers_list[0]);
+			if (!p)
+				return log_error_errno(-ENOENT, ENOENT, "Failed to retrieve limit cgroup path for %s", controllers_list[0] ?: "(null)");
 
-			path = strdup(p);
-			if (!path) {
-				ERROR("strdup failed");
-				goto err;
-			}
+			cgroup_base_path = strdup(p);
+			if (!cgroup_base_path)
+				return log_error_errno(-ENOMEM, ENOMEM, "Failed to duplicate limit cgroup path");
 		}
 
-		tmp = lxc_deslashify(path);
-		if (!tmp) {
-			ERROR("Failed to remove extraneous slashes from \"%s\"",
-			      path);
-			free(path);
-			goto err;
-		}
-		free(path);
-		path = tmp;
+		tmp = path_simplify(cgroup_base_path);
+		if (!tmp)
+			return log_error_errno(-ENOMEM, ENOMEM, "Failed to remove extraneous slashes from \"%s\"", tmp);
+		free_move_ptr(cgroup_base_path, tmp);
 
-		fullname = lxc_string_join(",", (const char **) controllers, false);
-		if (!fullname) {
-			ERROR("failed to join controllers");
-			free(path);
-			goto err;
-		}
+		if (controllers_list[0]) {
+			controllers = lxc_string_join(",", (const char **)controllers_list, false);
+			if (!controllers)
+				return log_error_errno(-ENOMEM, ENOMEM, "Failed to join controllers");
 
-		ret = sprintf(buf, "%s:%s", fullname, path);
-		free(path);
-		free(fullname);
-		if (ret < 0 || ret >= sizeof(buf)) {
-			ERROR("sprintf of cgroup root arg failed");
-			goto err;
+			ret = sprintf(buf, "%s:%s", controllers, cgroup_base_path);
+		} else {
+			WARN("No cgroup controllers configured in container's cgroup %s", cgroup_base_path);
+			ret = sprintf(buf, "%s", cgroup_base_path);
 		}
+		if (ret < 0 || ret >= sizeof(buf))
+			return log_error_errno(-EIO, EIO, "sprintf of cgroup root arg failed");
 
 		DECLARE_ARG("--cgroup-root");
 		DECLARE_ARG(buf);
@@ -351,65 +349,65 @@ static void exec_criu(struct cgroup_ops *cgroup_ops, struct lxc_conf *conf,
 		DECLARE_ARG(opts->user->action_script);
 	}
 
-	mnts = make_anonymous_mount_file(&opts->c->lxc_conf->mount_list,
+	f_mnt = make_anonymous_mount_file(&opts->c->lxc_conf->mount_list,
 	                                 opts->c->lxc_conf->lsm_aa_allow_nesting);
-	if (!mnts)
-		goto err;
+	if (!f_mnt)
+		return log_error_errno(-ENOENT, ENOENT, "Failed to create anonymous mount file");
 
-	while (getmntent_r(mnts, &mntent, buf, sizeof(buf))) {
+	while (getmntent_r(f_mnt, &mntent, buf, sizeof(buf))) {
+		__do_free char *mnt_options = NULL;
 		unsigned long flags = 0;
-		char *mntdata = NULL;
 		char arg[2 * PATH_MAX + 2];
 
-		if (parse_mntopts(mntent.mnt_opts, &flags, &mntdata) < 0)
-			goto err;
-
-		free(mntdata);
+		if (parse_mntopts_legacy(mntent.mnt_opts, &flags, &mnt_options) < 0)
+			return log_error_errno(-EINVAL, EINVAL, "Failed to parse mount options");
 
 		/* only add --ext-mount-map for actual bind mounts */
 		if (!(flags & MS_BIND))
 			continue;
 
-		if (strcmp(opts->action, "dump") == 0)
-			ret = snprintf(arg, sizeof(arg), "/%s:%s",
-				       mntent.mnt_dir, mntent.mnt_dir);
+		if (strequal(opts->action, "dump"))
+			ret = strnprintf(arg, sizeof(arg), "/%s:%s", mntent.mnt_dir, mntent.mnt_dir);
 		else
-			ret = snprintf(arg, sizeof(arg), "%s:%s",
-				       mntent.mnt_dir, mntent.mnt_fsname);
-		if (ret < 0 || ret >= sizeof(arg)) {
-			fclose(mnts);
-			ERROR("snprintf failed");
-			goto err;
-		}
+			ret = strnprintf(arg, sizeof(arg), "%s:%s", mntent.mnt_dir, mntent.mnt_fsname);
+		if (ret < 0)
+			return log_error_errno(-EIO, EIO, "Failed to create mount entry");
 
 		DECLARE_ARG("--ext-mount-map");
 		DECLARE_ARG(arg);
 	}
-	fclose(mnts);
 
-	if (strcmp(opts->action, "dump") == 0 || strcmp(opts->action, "pre-dump") == 0) {
-		char pid[32], *freezer_relative;
+	if (strequal(opts->action, "dump") || strequal(opts->action, "pre-dump")) {
+		pid_t init_pid;
+		char init_pid_str[INTTYPE_TO_STRLEN(int)];
+		char *freezer_relative;
 
-		if (sprintf(pid, "%d", opts->c->init_pid(opts->c)) < 0)
-			goto err;
+		init_pid = opts->c->init_pid(opts->c);
+		if (init_pid < 0)
+			return log_error_errno(-ESRCH, ESRCH, "Failed to retrieve init pid of container");
+
+		ret = strnprintf(init_pid_str, sizeof(init_pid_str), "%d", init_pid);
+		if (ret < 0)
+			return log_error_errno(-EIO, EIO, "Failed to create entry for init pid of container");
 
 		DECLARE_ARG("-t");
-		DECLARE_ARG(pid);
+		DECLARE_ARG(init_pid_str);
 
-		freezer_relative = lxc_cmd_get_limiting_cgroup_path(opts->c->name,
-								    opts->c->config_path,
-								    "freezer");
-		if (!freezer_relative) {
-			ERROR("failed getting freezer path");
-			goto err;
-		}
+		freezer_relative = lxc_cmd_get_limit_cgroup_path(opts->c->name,
+								 opts->c->config_path,
+								 "freezer");
+		if (!freezer_relative)
+			return log_error_errno(-ENOENT, ENOENT, "Failed getting freezer path");
 
-		ret = snprintf(log, sizeof(log), "/sys/fs/cgroup/freezer/%s", freezer_relative);
-		if (ret < 0 || ret >= sizeof(log))
-			goto err;
+		if (pure_unified_layout(cgroup_ops))
+			ret = strnprintf(log, sizeof(log), "/sys/fs/cgroup/%s", freezer_relative);
+		else
+			ret = strnprintf(log, sizeof(log), "/sys/fs/cgroup/freezer/%s", freezer_relative);
+		if (ret < 0)
+			return log_error_errno(-EIO, EIO, "Failed to freezer cgroup entry");
 
 		if (!opts->user->disable_skip_in_flight &&
-				strcmp(opts->criu_version, CRIU_IN_FLIGHT_SUPPORT) >= 0)
+		    strcmp(opts->criu_version, CRIU_IN_FLIGHT_SUPPORT) >= 0)
 			DECLARE_ARG("--skip-in-flight");
 
 		DECLARE_ARG("--freeze-cgroup");
@@ -444,21 +442,17 @@ static void exec_criu(struct cgroup_ops *cgroup_ops, struct lxc_conf *conf,
 			char ghost_limit[32];
 
 			ret = sprintf(ghost_limit, "%"PRIu64, opts->user->ghost_limit);
-			if (ret < 0 || ret >= sizeof(ghost_limit)) {
-				ERROR("failed to print ghost limit %"PRIu64, opts->user->ghost_limit);
-				goto err;
-			}
+			if (ret < 0 || ret >= sizeof(ghost_limit))
+				return log_error_errno(-EIO, EIO, "Failed to print ghost limit %"PRIu64, opts->user->ghost_limit);
 
 			DECLARE_ARG("--ghost-limit");
 			DECLARE_ARG(ghost_limit);
 		}
 
 		/* only for final dump */
-		if (strcmp(opts->action, "dump") == 0 && !opts->user->stop)
+		if (strequal(opts->action, "dump") && !opts->user->stop)
 			DECLARE_ARG("--leave-running");
-	} else if (strcmp(opts->action, "restore") == 0) {
-		void *m;
-		int additional;
+	} else if (strequal(opts->action, "restore")) {
 		struct lxc_conf *lxc_conf = opts->c->lxc_conf;
 
 		DECLARE_ARG("--root");
@@ -467,22 +461,20 @@ static void exec_criu(struct cgroup_ops *cgroup_ops, struct lxc_conf *conf,
 		DECLARE_ARG("--restore-sibling");
 
 		if (ttys[0]) {
-			if (opts->console_fd < 0) {
-				ERROR("lxc.console.path configured on source host but not target");
-				goto err;
-			}
+			if (opts->console_fd < 0)
+				return log_error_errno(-EINVAL, EINVAL, "lxc.console.path configured on source host but not target");
 
-			ret = snprintf(buf, sizeof(buf), "fd[%d]:%s", opts->console_fd, ttys);
-			if (ret < 0 || ret >= sizeof(buf))
-				goto err;
+			ret = strnprintf(buf, sizeof(buf), "fd[%d]:%s", opts->console_fd, ttys);
+			if (ret < 0)
+				return log_error_errno(-EIO, EIO, "Failed to create console entry");
 
 			DECLARE_ARG("--inherit-fd");
 			DECLARE_ARG(buf);
 		}
 		if (opts->console_name) {
-			if (snprintf(buf, sizeof(buf), "console:%s", opts->console_name) < 0) {
-				SYSERROR("sprintf'd too many bytes");
-			}
+			if (strnprintf(buf, sizeof(buf), "console:%s", opts->console_name) < 0)
+				return log_error_errno(-EIO, EIO, "Failed to create console entry");
+
 			DECLARE_ARG("--ext-mount-map");
 			DECLARE_ARG(buf);
 		}
@@ -490,24 +482,16 @@ static void exec_criu(struct cgroup_ops *cgroup_ops, struct lxc_conf *conf,
 		if (lxc_conf->lsm_aa_profile || lxc_conf->lsm_se_context) {
 
 			if (lxc_conf->lsm_aa_profile)
-				ret = snprintf(buf, sizeof(buf), "apparmor:%s", lxc_conf->lsm_aa_profile);
+				ret = strnprintf(buf, sizeof(buf), "apparmor:%s", lxc_conf->lsm_aa_profile);
 			else
-				ret = snprintf(buf, sizeof(buf), "selinux:%s", lxc_conf->lsm_se_context);
-
-			if (ret < 0 || ret >= sizeof(buf))
-				goto err;
+				ret = strnprintf(buf, sizeof(buf), "selinux:%s", lxc_conf->lsm_se_context);
+			if (ret < 0)
+				return log_error_errno(-EIO, EIO, "Failed to create lsm entry");
 
 			DECLARE_ARG("--lsm-profile");
 			DECLARE_ARG(buf);
 		}
 
-		additional = lxc_list_len(&opts->c->lxc_conf->network) * 2;
-
-		m = realloc(argv, (argc + additional + 1) * sizeof(*argv));
-		if (!m)
-			goto err;
-		argv = m;
-
 		lxc_list_for_each(it, &opts->c->lxc_conf->network) {
 			size_t retlen;
 			char eth[128], *veth;
@@ -527,11 +511,11 @@ static void exec_criu(struct cgroup_ops *cgroup_ops, struct lxc_conf *conf,
 			if (n->name[0] != '\0') {
 				retlen = strlcpy(eth, n->name, sizeof(eth));
 				if (retlen >= sizeof(eth))
-					goto err;
+					return log_error_errno(-E2BIG, E2BIG, "Failed to append veth device name");
 			} else {
-				ret = snprintf(eth, sizeof(eth), "eth%d", netnr);
-				if (ret < 0 || ret >= sizeof(eth))
-					goto err;
+				ret = strnprintf(eth, sizeof(eth), "eth%d", netnr);
+				if (ret < 0)
+					return log_error_errno(-E2BIG, E2BIG, "Failed to append veth device name");
 			}
 
 			switch (n->type) {
@@ -542,44 +526,37 @@ static void exec_criu(struct cgroup_ops *cgroup_ops, struct lxc_conf *conf,
 
 				if (n->link[0] != '\0') {
 					if (external_not_veth)
-						ret = snprintf(buf, sizeof(buf),
-							       "veth[%s]:%s@%s",
-							       eth, veth,
-							       n->link);
+						ret = strnprintf(buf, sizeof(buf), "veth[%s]:%s@%s", eth, veth, n->link);
 					else
-						ret = snprintf(buf, sizeof(buf),
-							       "%s=%s@%s", eth,
-							       veth, n->link);
+						ret = strnprintf(buf, sizeof(buf), "%s=%s@%s", eth, veth, n->link);
 				} else {
 					if (external_not_veth)
-						ret = snprintf(buf, sizeof(buf),
-							       "veth[%s]:%s",
-							       eth, veth);
+						ret = strnprintf(buf, sizeof(buf), "veth[%s]:%s", eth, veth);
 					else
-						ret = snprintf(buf, sizeof(buf),
-							       "%s=%s", eth,
-							       veth);
+						ret = strnprintf(buf, sizeof(buf), "%s=%s", eth, veth);
 				}
-				if (ret < 0 || ret >= sizeof(buf))
-					goto err;
+				if (ret < 0)
+					return log_error_errno(-EIO, EIO, "Failed to append veth device name");
+
+				TRACE("Added veth device entry %s", buf);
 				break;
 			case LXC_NET_MACVLAN:
-				if (n->link[0] == '\0') {
-					ERROR("no host interface for macvlan %s", n->name);
-					goto err;
-				}
+				if (n->link[0] == '\0')
+					return log_error_errno(-EINVAL, EINVAL, "Failed to find host interface for macvlan %s", n->name);
+
+				ret = strnprintf(buf, sizeof(buf), "macvlan[%s]:%s", eth, n->link);
+				if (ret < 0)
+					return log_error_errno(-EIO, EIO, "Failed to add macvlan entry");
+
+				TRACE("Added macvlan device entry %s", buf);
 
-				ret = snprintf(buf, sizeof(buf), "macvlan[%s]:%s", eth, n->link);
-				if (ret < 0 || ret >= sizeof(buf))
-					goto err;
 				break;
 			case LXC_NET_NONE:
 			case LXC_NET_EMPTY:
 				break;
 			default:
 				/* we have screened for this earlier... */
-				ERROR("unexpected network type %d", n->type);
-				goto err;
+				return log_error_errno(-EINVAL, EINVAL, "Unsupported network type %d", n->type);
 			}
 
 			if (external_not_veth)
@@ -592,42 +569,35 @@ static void exec_criu(struct cgroup_ops *cgroup_ops, struct lxc_conf *conf,
 
 	}
 
-	argv[argc] = NULL;
+	args->argv[args->argc] = NULL;
 
-	buf[0] = 0;
-	pos = 0;
+	if (lxc_log_trace()) {
+		buf[0] = 0;
+		for (int i = 0, pos = 0; i < args->argc && args->argv[i]; i++) {
+			ret = strnprintf(buf + pos, sizeof(buf) - pos, "%s ", args->argv[i]);
+			if (ret < 0)
+				return log_error_errno(-EIO, EIO, "Failed to reorder entries");
+			else
+				pos += ret;
+		}
 
-	for (i = 0; argv[i]; i++) {
-		ret = snprintf(buf + pos, sizeof(buf) - pos, "%s ", argv[i]);
-		if (ret < 0 || ret >= sizeof(buf) - pos)
-			goto err;
-		else
-			pos += ret;
+		TRACE("Using command line %s", buf);
 	}
 
-	INFO("execing: %s", buf);
-
 	/* before criu inits its log, it sometimes prints things to stdout/err;
 	 * let's be sure we capture that.
 	 */
-	if (dup2(opts->pipefd, STDOUT_FILENO) < 0) {
-		SYSERROR("dup2 stdout failed");
-		goto err;
-	}
+	if (dup2(opts->pipefd, STDOUT_FILENO) < 0)
+		return log_error_errno(-errno, errno, "Failed to duplicate stdout");
 
-	if (dup2(opts->pipefd, STDERR_FILENO) < 0) {
-		SYSERROR("dup2 stderr failed");
-		goto err;
-	}
+	if (dup2(opts->pipefd, STDERR_FILENO) < 0)
+		return log_error_errno(-errno, errno, "Failed to duplicate stderr");
 
 	close(opts->pipefd);
 
 #undef DECLARE_ARG
-	execv(argv[0], argv);
-err:
-	for (i = 0; argv[i]; i++)
-		free(argv[i]);
-	free(argv);
+	execv(args->argv[0], args->argv);
+	return -ENOEXEC;
 }
 
 /*
@@ -887,8 +857,8 @@ static bool restore_net_info(struct lxc_container *c)
 		if (netdev->type != LXC_NET_VETH)
 			continue;
 
-		ret = snprintf(template, sizeof(template), "vethXXXXXX");
-		if (ret < 0 || ret >= sizeof(template))
+		ret = strnprintf(template, sizeof(template), "vethXXXXXX");
+		if (ret < 0)
 			goto out_unlock;
 
 		if (netdev->priv.veth_attr.pair[0] == '\0' &&
@@ -938,14 +908,25 @@ static void do_restore(struct lxc_container *c, int status_pipe, struct migrate_
 
 	if (lxc_init(c->name, handler) < 0)
 		goto out;
+	cgroup_ops = handler->cgroup_ops;
 
-	cgroup_ops = cgroup_init(c->lxc_conf);
-	if (!cgroup_ops)
+	if (!cgroup_ops->monitor_create(cgroup_ops, handler)) {
+		ERROR("Failed to create monitor cgroup");
 		goto out_fini_handler;
-	handler->cgroup_ops = cgroup_ops;
+	}
+
+	if (!cgroup_ops->monitor_enter(cgroup_ops, handler)) {
+		ERROR("Failed to enter monitor cgroup");
+		goto out_fini_handler;
+	}
+
+	if (!cgroup_ops->monitor_delegate_controllers(cgroup_ops)) {
+		ERROR("Failed to delegate controllers to monitor cgroup");
+		goto out_fini_handler;
+	}
 
 	if (!cgroup_ops->payload_create(cgroup_ops, handler)) {
-		ERROR("failed creating groups");
+		ERROR("Failed creating cgroups");
 		goto out_fini_handler;
 	}
 
@@ -983,6 +964,10 @@ static void do_restore(struct lxc_container *c, int status_pipe, struct migrate_
 		if (unshare(CLONE_NEWNS))
 			goto out_fini_handler;
 
+		ret = lxc_storage_prepare(c->lxc_conf);
+		if (ret)
+			goto out_fini_handler;
+
 		/* CRIU needs the lxc root bind mounted so that it is the root of some
 		 * mount. */
 		rootfs = &c->lxc_conf->rootfs;
@@ -1034,7 +1019,9 @@ static void do_restore(struct lxc_container *c, int status_pipe, struct migrate_
 		os.console_name = c->lxc_conf->console.name;
 
 		/* exec_criu() returning is an error */
-		exec_criu(cgroup_ops, c->lxc_conf, &os);
+		ret = exec_criu(handler->cgroup_ops, c->lxc_conf, &os);
+		if (ret)
+			SYSERROR("Failed to execute criu");
 		umount(rootfs->mount);
 		(void)rmdir(rootfs->mount);
 		goto out_fini_handler;
@@ -1069,9 +1056,9 @@ static void do_restore(struct lxc_container *c, int status_pipe, struct migrate_
 				ERROR("criu process exited %d, output:\n%s", WEXITSTATUS(status), buf);
 				goto out_fini_handler;
 			} else {
-				ret = snprintf(buf, sizeof(buf), "/proc/self/task/%lu/children", (unsigned long)syscall(__NR_gettid));
-				if (ret < 0 || ret >= sizeof(buf)) {
-					ERROR("snprintf'd too many characters: %d", ret);
+				ret = strnprintf(buf, sizeof(buf), "/proc/self/task/%lu/children", (unsigned long)syscall(__NR_gettid));
+				if (ret < 0) {
+					ERROR("strnprintf'd too many characters: %d", ret);
 					goto out_fini_handler;
 				}
 
@@ -1114,8 +1101,8 @@ static void do_restore(struct lxc_container *c, int status_pipe, struct migrate_
 		 * fail because it's just a beauty thing. We just
 		 * assign the return here to silence potential.
 		 */
-		ret = snprintf(title, sizeof(title), "[lxc monitor] %s %s", c->config_path, c->name);
-		if (ret < 0 || (size_t)ret >= sizeof(title))
+		ret = strnprintf(title, sizeof(title), "[lxc monitor] %s %s", c->config_path, c->name);
+		if (ret < 0)
 			INFO("Setting truncated process name");
 
 		ret = setproctitle(title);
@@ -1161,14 +1148,14 @@ static int save_tty_major_minor(char *directory, struct lxc_container *c, char *
 	int ret;
 	struct stat sb;
 
-	if (c->lxc_conf->console.path && !strcmp(c->lxc_conf->console.path, "none")) {
+	if (c->lxc_conf->console.path && strequal(c->lxc_conf->console.path, "none")) {
 		tty_id[0] = 0;
 		return 0;
 	}
 
-	ret = snprintf(path, sizeof(path), "/proc/%d/root/dev/console", c->init_pid(c));
-	if (ret < 0 || ret >= sizeof(path)) {
-		ERROR("snprintf'd too many characters: %d", ret);
+	ret = strnprintf(path, sizeof(path), "/proc/%d/root/dev/console", c->init_pid(c));
+	if (ret < 0) {
+		ERROR("strnprintf'd too many characters: %d", ret);
 		return -1;
 	}
 
@@ -1178,17 +1165,17 @@ static int save_tty_major_minor(char *directory, struct lxc_container *c, char *
 		return -1;
 	}
 
-	ret = snprintf(path, sizeof(path), "%s/tty.info", directory);
-	if (ret < 0 || ret >= sizeof(path)) {
-		ERROR("snprintf'd too many characters: %d", ret);
+	ret = strnprintf(path, sizeof(path), "%s/tty.info", directory);
+	if (ret < 0) {
+		ERROR("strnprintf'd too many characters: %d", ret);
 		return -1;
 	}
 
-	ret = snprintf(tty_id, len, "tty[%llx:%llx]",
+	ret = strnprintf(tty_id, len, "tty[%llx:%llx]",
 					(long long unsigned) sb.st_rdev,
 					(long long unsigned) sb.st_dev);
-	if (ret < 0 || ret >= sizeof(path)) {
-		ERROR("snprintf'd too many characters: %d", ret);
+	if (ret < 0) {
+		ERROR("strnprintf'd too many characters: %d", ret);
 		return -1;
 	}
 
@@ -1259,7 +1246,9 @@ static bool do_dump(struct lxc_container *c, char *mode, struct migrate_opts *op
 		}
 
 		/* exec_criu() returning is an error */
-		exec_criu(cgroup_ops, c->lxc_conf, &os);
+		ret = exec_criu(cgroup_ops, c->lxc_conf, &os);
+		if (ret)
+			SYSERROR("Failed to execute criu");
 		free(criu_version);
 		_exit(EXIT_FAILURE);
 	} else {
@@ -1328,8 +1317,8 @@ bool __criu_dump(struct lxc_container *c, struct migrate_opts *opts)
 	char path[PATH_MAX];
 	int ret;
 
-	ret = snprintf(path, sizeof(path), "%s/inventory.img", opts->directory);
-	if (ret < 0 || ret >= sizeof(path))
+	ret = strnprintf(path, sizeof(path), "%s/inventory.img", opts->directory);
+	if (ret < 0)
 		return false;
 
 	if (access(path, F_OK) == 0) {
