@@ -13,6 +13,8 @@
 #include <unistd.h>
 
 #include "compiler.h"
+#include "memory_utils.h"
+#include "syscall_wrappers.h"
 
 /* read and write whole files */
 __hidden extern int lxc_write_to_file(const char *filename, const void *buf, size_t count,
@@ -73,12 +75,49 @@ static inline int fd_to_fd(int from, int to)
 {
 	return __fd_to_fd(from, to) >= 0;
 }
+__hidden extern int fd_cloexec(int fd, bool cloexec);
 __hidden extern int lxc_open_dirfd(const char *dir);
 __hidden extern FILE *fdopen_cached(int fd, const char *mode, void **caller_freed_buffer);
+__hidden extern FILE *fdopen_at(int dfd, const char *path, const char *mode,
+                                unsigned int o_flags,
+                                unsigned int resolve_flags);
 __hidden extern FILE *fopen_cached(const char *path, const char *mode, void **caller_freed_buffer);
+__hidden extern int timens_offset_write(clockid_t clk_id, int64_t s_offset, int64_t ns_offset);
 __hidden extern bool exists_dir_at(int dir_fd, const char *path);
 __hidden extern bool exists_file_at(int dir_fd, const char *path);
-__hidden extern int open_beneath(int dir_fd, const char *path, unsigned int flags);
+__hidden extern int open_at(int dfd, const char *path, unsigned int o_flags,
+			    unsigned int resolve_flags, mode_t mode);
+static inline int open_beneath(int dfd, const char *path, unsigned int flags)
+{
+	return open_at(dfd, path, flags, PROTECT_LOOKUP_BENEATH, 0);
+}
 __hidden int fd_make_nonblocking(int fd);
+__hidden extern char *read_file_at(int dfd, const char *fnam,
+                                   unsigned int o_flags,
+                                   unsigned resolve_flags);
+__hidden extern ssize_t lxc_read_try_buf_at(int dfd, const char *path,
+                                            void *buf, size_t count);
+
+/*
+ * Check if two fds refer to the same file.
+ * The function is "lax" in so far, as it doesn't care whether fda and fdb have
+ * the same flags or whether they share the same device context when they refer
+ * to devices.
+ */
+__hidden extern bool same_file_lax(int fda, int fdb);
+
+static inline int dup_cloexec(int fd)
+{
+	__do_close int fd_dup = -EBADF;
+
+	fd_dup = dup(fd);
+	if (fd_dup < 0)
+		return -errno;
+
+	if (fd_cloexec(fd_dup, true))
+		return -errno;
+
+	return move_fd(fd_dup);
+}
 
 #endif /* __LXC_FILE_UTILS_H */
