@@ -11,6 +11,7 @@
 #include "log.h"
 #include "macro.h"
 #include "memory_utils.h"
+#include "mount_utils.h"
 #include "storage.h"
 #include "utils.h"
 
@@ -126,44 +127,70 @@ bool dir_detect(const char *path)
 
 int dir_mount(struct lxc_storage *bdev)
 {
-	__do_free char *mntdata = NULL;
-	unsigned long mflags = 0, mntflags = 0, pflags = 0;
+	struct lxc_rootfs *rootfs = bdev->rootfs;
+	struct lxc_mount_options *mnt_opts = &rootfs->mnt_opts;
 	int ret;
-	const char *src;
+	const char *source, *target;
 
-	if (strcmp(bdev->type, "dir"))
-		return -22;
+	if (!strequal(bdev->type, "dir"))
+		return syserror_set(-EINVAL, "Invalid storage driver");
 
-	if (!bdev->src || !bdev->dest)
-		return -22;
+	if (is_empty_string(bdev->src))
+		return syserror_set(-EINVAL, "Missing rootfs path");
 
-	ret = parse_mntopts(bdev->mntopts, &mntflags, &mntdata);
-	if (ret < 0)
-		return log_error_errno(ret, errno, "Failed to parse mount options \"%s\"", bdev->mntopts);
+	if (is_empty_string(bdev->dest))
+		return syserror_set(-EINVAL, "Missing target mountpoint");
 
-	ret = parse_propagationopts(bdev->mntopts, &pflags);
-	if (ret < 0)
-		return log_error_errno(-EINVAL, EINVAL, "Failed to parse mount propagation options \"%s\"", bdev->mntopts);
+	if (rootfs->dfd_idmapped >= 0 && !can_use_bind_mounts())
+		return syserror_set(-EOPNOTSUPP, "Idmapped mount requested but kernel doesn't support new mount API");
 
-	src = lxc_storage_get_path(bdev->src, bdev->type);
+	source = lxc_storage_get_path(bdev->src, bdev->type);
+	target = bdev->dest;
 
-	ret = mount(src, bdev->dest, "bind", MS_BIND | MS_REC | mntflags | pflags, mntdata);
-	if (ret < 0)
-		return log_error_errno(-errno, errno, "Failed to mount \"%s\" on \"%s\"", src, bdev->dest);
+	if (can_use_bind_mounts()) {
+		__do_close int fd_source = -EBADF, fd_target = -EBADF;
 
-	if (ret == 0 && (mntflags & MS_RDONLY)) {
-		mflags = add_required_remount_flags(src, bdev->dest, MS_BIND | MS_REC | mntflags | pflags | MS_REMOUNT);
-		ret = mount(src, bdev->dest, "bind", mflags, mntdata);
-		if (ret < 0)
-			return log_error_errno(-errno, errno, "Failed to remount \"%s\" on \"%s\" read-only with options \"%s\", mount flags \"%lu\", and propagation flags \"%lu\"",
-					       src ? src : "(none)", bdev->dest ? bdev->dest : "(none)", mntdata, mflags, pflags);
-		else
-			DEBUG("Remounted \"%s\" on \"%s\" read-only with options \"%s\", mount flags \"%lu\", and propagation flags \"%lu\"",
-			      src ? src : "(none)", bdev->dest ? bdev->dest : "(none)", mntdata, mflags, pflags);
+		fd_target = open_at(-EBADF, target, PROTECT_OPATH_DIRECTORY, 0, 0);
+		if (fd_target < 0)
+			return syserror("Failed to open \"%s\"", target);
+
+		if (rootfs->dfd_idmapped >= 0) {
+			ret = move_detached_mount(rootfs->dfd_idmapped, fd_target, "",
+						  PROTECT_OPATH_DIRECTORY,
+						  PROTECT_LOOKUP_BENEATH);
+		} else {
+			fd_source = open_at(-EBADF, source, PROTECT_OPATH_DIRECTORY, 0, 0);
+			if (fd_source < 0)
+				return syserror("Failed to open \"%s\"", source);
+
+			ret = fd_bind_mount(fd_source, "",
+					    PROTECT_OPATH_DIRECTORY,
+					    PROTECT_LOOKUP_BENEATH, fd_target,
+					    "", PROTECT_OPATH_DIRECTORY,
+					    PROTECT_LOOKUP_BENEATH, 0, true);
+		}
+	} else {
+		ret = mount(source, target, "bind", MS_BIND | MS_REC | mnt_opts->mnt_flags | mnt_opts->prop_flags, mnt_opts->data);
+		if (!ret && (mnt_opts->mnt_flags & MS_RDONLY)) {
+			unsigned long mflags;
+
+			mflags = add_required_remount_flags(source, target,
+							    MS_BIND |
+							    MS_REC |
+							    mnt_opts->mnt_flags |
+							    MS_REMOUNT);
+
+			ret = mount(source, target, "bind", mflags, mnt_opts->data);
+			if (ret)
+				SYSERROR("Failed to remount \"%s\" on \"%s\" read-only", source, target);
+			else
+				TRACE("Remounted \"%s\" on \"%s\" read-only", source, target);
+		}
 	}
+	if (ret < 0)
+		return syserror_set(ret, "Failed to mount \"%s\" onto \"%s\"", source, target);
 
-	TRACE("Mounted \"%s\" on \"%s\" with options \"%s\", mount flags \"%lu\", and propagation flags \"%lu\"",
-	      src ? src : "(none)", bdev->dest ? bdev->dest : "(none)", mntdata, mflags, pflags);
+	TRACE("Mounted \"%s\" onto \"%s\"", source, target);
 	return 0;
 }
 
