@@ -314,6 +314,14 @@ struct lxc_storage *storage_copy(struct lxc_container *c, const char *cname,
 	const char *oldpath = c->config_path;
 	char cmd_output[PATH_MAX] = {0};
 	struct rsync_data data = {0};
+	struct lxc_rootfs new_rootfs = {
+		.managed		= true,
+		.dfd_mnt		= -EBADF,
+		.dfd_dev		= -EBADF,
+		.dfd_host		= -EBADF,
+		.fd_path_pin		= -EBADF,
+		.dfd_idmapped		= -EBADF,
+	};
 
 	if (!src) {
 		ERROR("No rootfs specified");
@@ -329,11 +337,21 @@ struct lxc_storage *storage_copy(struct lxc_container *c, const char *cname,
 		return NULL;
 	}
 
-	orig = storage_init(c->lxc_conf);
-	if (!orig) {
+	ret = lxc_storage_prepare(c->lxc_conf);
+	if (ret) {
 		ERROR("Failed to detect storage driver for \"%s\"", oldname);
 		return NULL;
 	}
+	orig = c->lxc_conf->rootfs.storage;
+
+	if (c->lxc_conf->rootfs.dfd_idmapped >= 0) {
+		new_rootfs.dfd_idmapped = dup_cloexec(c->lxc_conf->rootfs.dfd_idmapped);
+		if (new_rootfs.dfd_idmapped < 0) {
+			SYSERROR("Failed to duplicate user namespace file descriptor");
+			lxc_storage_put(c->lxc_conf);
+			return NULL;
+		}
+	}
 
 	if (!orig->dest) {
 		size_t len;
@@ -404,6 +422,7 @@ struct lxc_storage *storage_copy(struct lxc_container *c, const char *cname,
 		goto on_error_put_orig;
 	}
 	TRACE("Initialized %s storage driver", new->type);
+	new->rootfs = &new_rootfs;
 
 	/* create new paths */
 	ret = new->ops->clone_paths(orig, new, oldname, cname, oldpath, lxcpath,
@@ -499,7 +518,7 @@ struct lxc_storage *storage_copy(struct lxc_container *c, const char *cname,
 	}
 
 on_success:
-	storage_put(orig);
+	lxc_storage_put(c->lxc_conf);
 
 	return new;
 
@@ -507,7 +526,7 @@ on_error_put_new:
 	storage_put(new);
 
 on_error_put_orig:
-	storage_put(orig);
+	lxc_storage_put(c->lxc_conf);
 
 	return NULL;
 }
@@ -598,14 +617,13 @@ struct lxc_storage *storage_init(struct lxc_conf *conf)
 	if (!q)
 		return NULL;
 
-	bdev = malloc(sizeof(struct lxc_storage));
+	bdev = zalloc(sizeof(struct lxc_storage));
 	if (!bdev)
 		return NULL;
 
-	memset(bdev, 0, sizeof(struct lxc_storage));
-
-	bdev->ops = q->ops;
-	bdev->type = q->name;
+	bdev->ops	= q->ops;
+	bdev->type	= q->name;
+	bdev->rootfs	= &conf->rootfs;
 
 	if (mntopts)
 		bdev->mntopts = strdup(mntopts);
@@ -644,10 +662,12 @@ bool storage_is_dir(struct lxc_conf *conf)
 
 void storage_put(struct lxc_storage *bdev)
 {
-	free(bdev->mntopts);
-	free(bdev->src);
-	free(bdev->dest);
-	free(bdev);
+	if (bdev) {
+		free_disarm(bdev->mntopts);
+		free_disarm(bdev->src);
+		free_disarm(bdev->dest);
+		free_disarm(bdev);
+	}
 }
 
 bool rootfs_is_blockdev(struct lxc_conf *conf)
