@@ -30,8 +30,10 @@
 #include <sys/stat.h>
 
 #include "cgroup.h"
-#include "lxc.h"
 #include "commands.h"
+#include "lxc.h"
+#include "lxctest.h"
+#include "utils.h"
 
 #ifndef HAVE_STRLCPY
 #include "include/strlcpy.h"
@@ -50,12 +52,31 @@
  */
 static int test_running_container(const char *lxcpath, const char *name)
 {
+	__do_close int fd_log = -EBADF;
 	int ret = -1;
 	struct lxc_container *c = NULL;
+	struct lxc_log log = {};
+	char template[sizeof(P_tmpdir"/attach_XXXXXX")];
 	char *cgrelpath;
 	char  relpath[PATH_MAX+1];
 	char  value[NAME_MAX], value_save[NAME_MAX];
-	struct cgroup_ops *cgroup_ops;
+	call_cleaner(cgroup_exit) struct cgroup_ops *cgroup_ops = NULL;
+
+	(void)strlcpy(template, P_tmpdir"/attach_XXXXXX", sizeof(template));
+
+	fd_log = lxc_make_tmpfile(template, false);
+	if (fd_log < 0) {
+		lxc_error("Failed to create temporary log file for container %s\n", name);
+		exit(EXIT_FAILURE);
+	}
+	log.name = name;
+	log.file = template;
+	log.level = "TRACE";
+	log.prefix = "cgpath";
+	log.quiet = false;
+	log.lxcpath = NULL;
+	if (lxc_log_init(&log))
+		goto err1;
 
 	sprintf(relpath, DEFAULT_PAYLOAD_CGROUP_PREFIX "%s", name);
 
@@ -83,33 +104,33 @@ static int test_running_container(const char *lxcpath, const char *name)
 		goto err3;
 
 	/* test get/set value using memory.soft_limit_in_bytes file */
-	ret = cgroup_ops->get(cgroup_ops, "memory.soft_limit_in_bytes", value,
-			      sizeof(value), c->name, c->config_path);
-	if (ret < 0) {
+	ret = cgroup_ops->get(cgroup_ops, "pids.max", value, sizeof(value),
+			      c->name, c->config_path);
+        if (ret < 0) {
 		TSTERR("cgroup_get failed");
 		goto err3;
 	}
 	(void)strlcpy(value_save, value, NAME_MAX);
 
-	ret = cgroup_ops->set(cgroup_ops, "memory.soft_limit_in_bytes", "512M",
+	ret = cgroup_ops->set(cgroup_ops, "pids.max", "10000",
 			      c->name, c->config_path);
 	if (ret < 0) {
 		TSTERR("cgroup_set failed %d %d", ret, errno);
 		goto err3;
 	}
-	ret = cgroup_ops->get(cgroup_ops, "memory.soft_limit_in_bytes", value,
+	ret = cgroup_ops->get(cgroup_ops, "pids.max", value,
 			      sizeof(value), c->name, c->config_path);
 	if (ret < 0) {
 		TSTERR("cgroup_get failed");
 		goto err3;
 	}
-	if (strcmp(value, "536870912\n")) {
+	if (strcmp(value, "10000\n")) {
 		TSTERR("cgroup_set_bypath failed to set value >%s<", value);
 		goto err3;
 	}
 
 	/* restore original value */
-	ret = cgroup_ops->set(cgroup_ops, "memory.soft_limit_in_bytes",
+	ret = cgroup_ops->set(cgroup_ops, "pids.max",
 			      value_save, c->name, c->config_path);
 	if (ret < 0) {
 		TSTERR("cgroup_set failed");
@@ -123,6 +144,17 @@ err3:
 err2:
 	lxc_container_put(c);
 err1:
+
+	if (ret != 0) {
+		char buf[4096];
+		ssize_t buflen;
+		while ((buflen = read(fd_log, buf, 1024)) > 0) {
+			buflen = write(STDERR_FILENO, buf, buflen);
+			if (buflen <= 0)
+				break;
+		}
+	}
+	(void)unlink(template);
 	return ret;
 }
 
@@ -148,6 +180,7 @@ static int test_container(const char *lxcpath, const char *name,
 	if (c->is_defined(c)) {
 		c->stop(c);
 		c->destroy(c);
+		lxc_container_put(c);
 		c = lxc_container_new(name, lxcpath);
 	}
 	c->set_config_item(c, "lxc.net.0.type", "empty");
@@ -173,7 +206,7 @@ out1:
 	return ret;
 }
 
-int main()
+int main(int argc, char *argv[])
 {
 	int ret = EXIT_FAILURE;
 
